#ifndef __DATAPOINT_H__
#define __DATAPOINT_H__

#include "../std.h"
#include <string>
#include <list>
#include "../transfer/url_map.h"
#include "../misc/globus_modules.h"

class RCManager;

#ifdef HAVE_GLOBUS_RLS_CLIENT_H
extern "C" {
#include <globus_rls_client.h>
}
#endif


/// DataPoint is an abstraction of URL. It can handle URLs of type file://, 
/// ftp://, gsiftp://, http://, https://, httpg:// (HTTP over GSI), 
/// se:// (NG web service over HTTPG) and meta-URLs (URLs of Infexing 
/// Services) rc://, rls://.
/// DataPoint provides means to resolve meta-URL into multiple URLs and
/// to loop through them.
class DataPoint {
  friend ostream& operator<<(ostream& o,const DataPoint &point);
 public:
  /// DataPoint::Location represent physical service at which files are 
  /// located aka "base URL" inculding it's name (as given in Indexing Service).
  /// Currently it is used only internally by DataPoint class and for 
  /// printing debug information. 
  class Location {
   friend class DataPoint;
   friend ostream& operator<<(ostream& o,const DataPoint &point);
   private:
    string meta; // Given name of location
    string url;  // location aka pfn aka access point
    bool existing;
    void* arg;   // to be used by different pieces of soft differently
   public:
    Location(void):existing(true),arg(NULL) { };
    Location(const char* url_):meta(""),url(url_),existing(true),arg(NULL) { };
    Location(const char* meta_,const char* url_,bool existing_ = true):meta(meta_),url(url_?url_:""),existing(existing_),arg(NULL) { };
    Location(const string &url_):meta(""),url(url_),existing(true),arg(NULL) { };
    Location(const string &meta_,const string &url_):meta(meta_),url(url_),existing(true),arg(NULL) { };
  };
  /// FileInfo stores information about file (meta-information). Although all 
  /// members are public it is mot desirable to modify them directly outside
  /// DataPoint class.
  class FileInfo {
   public:
    typedef enum {
      file_type_unknown = 0,
      file_type_file = 1,
      file_type_dir = 2
    } Type;
    string name;
    list<string> urls; /// Physical enpoints/URLs at which file can be accessed.
    unsigned long long int size;  /// Size of filein bytes.
    bool size_available;          /// If size is known.
    unsigned long long int checksum;  /// Checksum of file.
    bool checksum_available;          /// If checksum is known.
    time_t created;          /// Creation/modification time.
    bool created_available;  /// If time is known.
    time_t valid;            /// Valid till time.
    bool valid_available;    /// If validity is known.
    Type type;               /// Fyle type - usually file_type_file - ordinary file.
    FileInfo(const char *name_ = ""):name(name_),size_available(false),checksum_available(false),created_available(false),valid_available(false),type(file_type_unknown) {};
    operator bool(void) { return (name.length() != 0); }; /// If object is valid
  };
  list<Location> locations; /// List of locations at which file can be probably found.
 private:
  bool is_valid;
  string url;
  string common_url_options;
  list<Location>::iterator location;
  bool is_secure;
  bool is_resolved;
/*  bool globus_module_active; */

  bool is_metaexisting;

  unsigned long long int meta_size_;
  bool meta_size_valid;
  unsigned long long int meta_checksum_;
  bool meta_checksum_valid;
  time_t meta_created_;
  bool meta_created_valid;
  time_t meta_validtill_;
  bool meta_validtill_valid;
/*  bool meta_valid; */

  /* TODO - make it union */
  /* file */
  bool is_file;
  /* ftp */
  bool is_ftp;
  /* http */
  bool is_http;
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
  /* RC */
  string rc_url;
  RCManager *rc_mgr;
  string rc_lfn;
  bool is_rc;
#endif
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  /* rls */
  string rls_url;
  string rls_lfn;
  bool is_rls;
#endif
#ifndef SKIP_HTTPG
  /* httpg */
  bool is_httpg;
  /* ng se (ns) */
  bool is_se;
  string se_url;
  string se_lfn;
#endif
  /*  retries  */
  int tries_left;
  /* globus modules */
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  GlobusModuleCommon common_mod;
#endif
  GlobusModuleError error_mod;
  GlobusModuleGSIGSSAPI gsi_gssapi_mod;
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
  GlobusModuleReplicaCatalog replica_mod;
#endif
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  GlobusModuleIO io_mod;
  GlobusModuleRLSClient rls_mod;
#endif

  bool meta_resolve_rc(bool source);
  bool meta_preregister_rc(bool replication,bool force = false);
  bool meta_postregister_rc(bool replication,bool failure);
  bool meta_preunregister_rc(bool replication);
  bool meta_unregister_rc(bool all);
  bool process_meta_url_rc(const string &url,string &cat_url,string &filename,list<DataPoint::Location> &locations);
  bool list_files_rc(list<DataPoint::FileInfo> &files,bool resolve = true);

  bool meta_resolve_rls(bool source);
  bool meta_preregister_rls(bool replication,bool force = false);
  bool meta_postregister_rls(bool replication,bool failure);
  bool meta_preunregister_rls(bool replication);
  bool meta_unregister_rls(bool all);
  bool process_meta_url_rls(const string &url,string &cat_url,string &filename,list<DataPoint::Location> &locations);
  bool list_files_rls(list<DataPoint::FileInfo> &files,bool resolve = true);
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  static bool list_files_rls_callback(globus_rls_handle_t* h,const char* url,void* arg);
  static bool meta_resolve_rls_callback(globus_rls_handle_t* h,const char* url,void* arg);
#endif

#ifndef SKIP_HTTPG
  bool meta_resolve_se(bool source);
  bool meta_preregister_se(bool replication,bool force = false);
  bool meta_postregister_se(bool replication,bool failure);
  bool meta_preunregister_se(bool replication);
  bool meta_unregister_se(bool all);
  bool process_meta_url_se(const string &url,string &cat_url,string &filename,list<DataPoint::Location> &locations);
  bool list_files_se(list<DataPoint::FileInfo> &files,bool resolve = true);
#endif

 public:
  /// Constructor requres URL or meta-URL to be provided
  DataPoint(const char* url); 
  ~DataPoint(void);
  /// Resolve meta-URL into list of ordinary URLs and obtain meta-information 
  /// about file. Can be called for object representing ordinary URL or
  /// already resolved object.
  /// \param source true if DataPoint object represents source of information
  bool meta_resolve(bool source);
  /// Resolve meta-URL into list of ordinary URLs and obtain meta-information 
  /// about file. Also sort obtained list so that URLs mentioned in UrlMap
  /// object are placed first. This is used during transfer to access
  /// local locations first.
  /// \param maps list of mappings of remote URLs to (potentially) local locations.
  bool meta_resolve(bool source,const UrlMap &maps);
  ///  This function registers physical location of file into Indexing
  ///  Service. It should be called *before* actual transfer to that 
  ///  location happens.
  ///  \param replication if true then file is being replicated between 2 locations registered in Indexing Service under same name.
  ///  \param force if true, perform registration of new file even if it already exists. Should be used to fix failures in Indexing Service.
  bool meta_preregister(bool replication,bool force = false);
  /// Used for same purpose as meta_preregister. Should be called after
  /// actual transfer of file successfully finished.
  ///  \param replication if true then file is being replicated between 2 locations registered in Indexing Service under same name.
  /// \param failure not used.
  bool meta_postregister(bool replication,bool failure);
  ///  Should be called if file transfer failed. It removes changwes made 
  ///  by meta_preregister.
  bool meta_preunregister(bool replication);
  ///  Remove information about file registered in Indexing Service.
  ///  \param all if true information about file itself is (LFN) is removed. Otherwise only particular physical instance is unregistered.
  bool meta_unregister(bool all);
  /// Obtain information about objects and their properties available 
  /// under meta-URL of DataPoint object. It works only for meta-URL.
  /// \param files list of obtained objects.
  /// \param resolve if false, do not try to obtain propertiers of objects.
  bool list_files(list<DataPoint::FileInfo> &files,bool resolve = true);
  /// Retrieve properties of object pointed by meta-URL of DataPoint 
  /// object. It works only for meta-URL. 
  /// \param fi contains retrieved information.
  bool get_info(DataPoint::FileInfo &fi);
  /*
    Set and get corresponding meta-information related to lfn.
  */
  /// Check if meta-information 'size' is available.
  bool meta_size_available(void) { return meta_size_valid; };
  /// Set value of meta-information 'size' if not already set.
  void meta_size(unsigned long long int val) {
    if(!meta_size_valid) { meta_size_=val; meta_size_valid=true; };
  };
  /// Set value of meta-information 'size'.
  void meta_size_force(unsigned long long int val) {
    meta_size_=val; meta_size_valid=true;
  };
  /// Get value of meta-information 'size'.
  unsigned long long int meta_size(void) const {
    if(meta_size_valid) return meta_size_; return 0;
  };
  /// Check if meta-information 'checksum' is available.
  bool meta_checksum_available(void) { return meta_checksum_valid; };
  /// Set value of meta-information 'checksum' if not already set.
  void meta_checksum(unsigned long long int val) {
    if(!meta_checksum_valid) { meta_checksum_=val; meta_checksum_valid=true; };
  };
  /// Set value of meta-information 'checksum'.
  void meta_checksum_force(unsigned long long int val) {
    meta_checksum_=val; meta_checksum_valid=true;
  };
  /// Get value of meta-information 'checksum'.
  unsigned long long int meta_checksum(void) const {
    if(meta_checksum_valid) return meta_checksum_; return 0;
  };
  /// Check if meta-information 'creation/modification time' is available.
  bool meta_created_available(void) { return meta_created_valid; };
  /// Set value of meta-information 'creation/modification time' if not already set.
  void meta_created(time_t val) {
    if(!meta_created_valid) { meta_created_=val; meta_created_valid=true; };
  };
  /// Set value of meta-information 'creation/modification time'.
  void meta_created_force(time_t val) {
    meta_created_=val; meta_created_valid=true;
  };
  /// Get value of meta-information 'creation/modification time'.
  time_t meta_created(void) const {
    if(meta_created_valid) return meta_created_; return 0;
  };
  /// Check if meta-information 'validity time' is available.
  bool meta_validtill_available(void) { return meta_validtill_valid; };
  /// Set value of meta-information 'validity time' if not already set.
  void meta_validtill(time_t val) {
    if(!meta_validtill_valid) {meta_validtill_=val; meta_validtill_valid=true;};
  };
  /// Set value of meta-information 'validity time'.
  void meta_validtill_force(time_t val) {
    meta_validtill_=val; meta_validtill_valid=true;
  };
  /// Get value of meta-information 'validity time'.
  time_t meta_validtill(void) const {
    if(meta_validtill_valid) return meta_validtill_; return 0;
  };
  /*
    return true if url is "meta-url" (rc://, rls://, etc.)
  */
  /// Check if URL is meta-URL.
  bool meta(void) const { return
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
                            is_rc ||
#endif
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
                            is_rls ||
#endif
/*
#ifndef SKIP_HTTPG
                            is_se ||
#endif
*/
                            false;
  };
  /// Acquire meta-information form another object. Defined values a 
  /// not overwritten.
  /// \param p object from which information is taken.
  void meta(const DataPoint &p) {
    if(p.meta_size_valid)      meta_size(p.meta_size_);
    if(p.meta_checksum_valid)  meta_checksum(p.meta_checksum_);
    if(p.meta_created_valid)   meta_created(p.meta_created_);
    if(p.meta_validtill_valid) meta_validtill(p.meta_validtill_);
  };
  /// Compare meta-information form another object. Undefined values
  /// are not used for comparison. Default result is 'true'. 
  /// \param p object to which compare.
  bool meta_compare(const DataPoint &p) {
    if(p.meta_size_valid && meta_size_valid)
      if(meta_size_ != p.meta_size_) return false;
    if(p.meta_checksum_valid && meta_checksum_valid)
      if(meta_checksum_ != p.meta_checksum_) return false;
    if(p.meta_created_valid && meta_created_valid)
      if(meta_created_ != p.meta_created_) return false;
    if(p.meta_validtill_valid && meta_validtill_valid)
      if(meta_validtill_ != p.meta_validtill_) return false;
    return true;
  };
  /// Check if file is registered in Indexing Service. Proper value is 
  /// obtinable only after meta-resolve.
  bool meta_stored(void) { return is_metaexisting; };
  /// Check if file is local (URL is file://).
  bool local(void) const { return is_file; };
  /// Map url (change it) according to table provided in maps.
  /// \param maps mapping information.
  bool map(const UrlMap &maps);
  /// Sort list of URLs so that those listed in mapping table are put first.
  /// \param maps mapping information.
  bool sort(const UrlMap &maps);
  DataPoint& operator= (const DataPoint &);
  operator bool (void) const { return is_valid; };
  /// Returns current (resolved) URL.
  const string& current_location(void) const { return location->url; };
  /// Returns meta information used to create curent URL. For RC that is 
  ///  location's name. For RLS that is equal to pfn.
  const string& current_meta_location(void) const { return location->meta; };
  /// Switch to next location in list of URLs. At last location 
  /// switch to first if number of allowed retries does s not exceeded.
  /// Returns false if no retries left.
  bool next_location(void);
  /// Returns false if out of retries. 
  bool have_location(void);
  /// Returns true if number of resolved URLs is not 0.
  bool have_locations(void);
  /// Remove remove current URL from list
  bool remove_location(void);
  /// Returns number of retries left.
  int tries(void) { return tries_left; };
  /// Set number of retries.
  void tries(int n) { tries_left=n; };
  /// Returns URL which was passed to constructor
  string base_url(void) const;
  /// Returns URL which was passed to constructor with location 
  /// names removed, port number added, etc.
  string canonic_url(void) const;
  /// Returns name which is given to file in Indexing Service (aka LFN).
  const char* lfn(void) const {
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
    if(is_rc) return rc_lfn.c_str();
#endif
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
    if(is_rls) return rls_lfn.c_str();
#endif
#ifndef SKIP_HTTPG
    if(is_se) return se_lfn.c_str();
#endif
    return "";
  };
  /// Add URL to list.
  /// \param meta meta-name (name of location/service).
  /// \param loc URL.
  bool add_location(const char* meta,const char* loc);
};

ostream& operator<<(ostream& o,const DataPoint &point);

#endif
