/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/
/*
 * This module contains the functions that execute queries on the VOMS database.
 * All access is done via the layer of indirection provided by the functions in
 * the mysql_wrap.c module.
 */
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <string.h>

#include "mysqlwrap.h"
#include "access_db_sql.h"
#include "data.h"

#if (__GNUC__ >= 3)
#include <ext/hash_set>
using  __gnu_cxx::hash_set;
#else
#include <hash_set>
#endif

typedef set<std::string> datahash;

/* access_db_sql.c */
static bool get_group_and_role_real(const std::string &dn, 
				    const std::string &ca, 
				    const std::string &dbname, 
				    const std::string &username, 
				    const char *password, 
				    const std::string &query, 
				    std::vector<attrib> &result);
static bool simple_query(const std::string &dbname, 
			 const std::string &username, 
			 const char *password, const std::string &query,
			 const char *fieldname, std::string &result);
static bool get_fields(const std::string &dbname, const std::string &username,
		       const char *password, const std::string &query, 
		       std::string &result);
#if MYSQL_CORRECTED
static bool get_group_and_role_real0(sql::interface &db, const std::string &dn,
				     const std::string &ca,
				     const std::string &query,
				     std::vector<attrib> &results, datahash &map);
#else
static bool get_group_and_role_real0(sql::interface &db, sql::interface &db2,
				     const std::string &dn, const std::string &ca,
				     const std::string &query, 
				     std::vector<attrib> &results, datahash &map);
#endif
static bool simple_query0(sql::interface &db, const std::string &query, const char *fieldname, std::string &result);

/*
 * Function:
 *   subst(format, user, ca)
 *
 * Description:
 *   This function inserts into the 'format' std::string the values of 'user' and
 *   'ca'.
 *
 * Parameters:
 *   'format' - This is the prototype std::string. It may contain the substd::strings
 *              '$user' and '$ca', that are substituted respectively with the
 *              content of the 'user' and 'ca' arguments.
 *   'user'   - This is the value to substitute in place of the '$user' std::string.
 *   'ca'     - This is the value to substitute in place of the '$ca' std::string.
 *
 * Result:
 *   Failure:
 *     NULL.
 *   Success:
 *     A pointer to a new std::string with the values substituted. It must be manually
 *     freed when it is no longer used.
 *
 * NOTE:
 *   All the substitutions are performed at the same time.
 */
static bool
subst(std::string &format, const std::string &user, const std::string &ca)
{
  int doca;

  if (format.empty() || user.empty() || ca.empty())
    return false;

  /*
   * Inefficient, but will do for now.
   */

  unsigned int userp = format.find("$user");
  unsigned int cap = format.find("$ca");

  while (userp != format.npos && cap != format.npos) {

    /*
     * Selects the first substitution to perform.
     */
    if (cap < userp)
      doca = 1;
    else
      doca = 0;
    if (!cap)
      doca = 0;
    if (!userp)
      doca = 1;

    /*
     * Substitutes an element.
     */
    switch (doca) {
    case 0:
      format.replace(userp, 5, user);
      break;

    case 1:
      format.replace(cap, 3, ca);
      break;

    default:
      /* should never happen */
      return false;
    }

    userp = format.find("$user");
    cap = format.find("$ca");
  }

  return true;
}

/*
 * Function:
 *   special(dn, ca, request, dbname, username, password)
 *
 * Description:
 *   This function retireves a query from the query database and then executes
 *   it.
 *
 * Parameters:
 *   'dn'       - The subject of the user's certificate.
 *   'ca'       - The subject of the CA that issued the user's certificate.
 *   'request'  - The request that must be performed. In the format "S<num>" where
 *                the "S" part is supposed to having been verified by the caller.
 *   'dbname'   - The name of the VOMS database.
 *   'username' - The username of the user that must access the VOMS database.
 *   'password' - The password of the user that must access the VOMS database.
 *
 * Result:
 *   Failure:
 *     NULL. Furthermore, the 'data' blob is erased.
 *   Success:
 *     The blob with the requested data.
 *
 * NOTE:
 *   The data is returned as a list of "<fieldname>: <value>" lines for each
 *   field. Multiple records are added one after another.
 */
bool
special(const std::string &dn, const std::string &ca, const char *request, 
	const std::string &dbname, const std::string &username, const char *password,
	std::string &result)
{
  if (dn.empty() || ca.empty() || !request || dbname.empty() ||
      username.empty() || !password)
    return false;

  /* Verifies if the input is acceptable */
  if (acceptable(request)) {
    std::string query  = std::string("SELECT query FROM queries WHERE qid = ") + request;
    std::string realquery;

    /* Retrieves the query */
    if (simple_query(dbname, username, password, query, "query", realquery))
      if (subst(realquery, dn, ca))
	return get_fields(dbname, username, password, realquery, result);
  }

  return false;
}


/*
 * Function:
 *   listspecial(dbname, username, password)
 *
 * Description:
 *   This function returns a blob containing the text of all the special queries.
 *
 * Parameters:
 *   'dbname'   - The name of the VOMS database.
 *   'username' - The username of the user that must access the VOMS database.
 *   'password' - The password of the user that must access the VOMS database.
 *
 * Result:
 *   Failure:
 *     NULL. Furthermore, the 'data' blob is erased.
 *   Success:
 *     The blob with the requested data.
 */
bool
listspecial(const std::string &dbname, const std::string &username,
	    const char *password, std::string &result)
{
  return get_fields(dbname, username, password, "SELECT qid, query FROM queries", result);
}

/*
 * Function:
 *   get_all(dn, ca, dbname, username, password)
 *
 * Description:
 *   This function creates a SQL query that retrieves informations about
 *   the groups and roles to which the user belongs.
 *
 * Parameters:
 *   'username' - The username used for database login.
 *   'password' - The password used for database login.
 *   'dbname'   - The database name.
 *   'dn'       - Distinguished Name of the user's certificate.
 *   'ca'       - Distinguished Name of the CA that issued the user's
 *                certificate.
 *
 * Return:
 *   Success - 
 *     The blob with the requested data.
 *   Failure -
 *     NULL.
 */
bool
get_all(const std::string &dn, const std::string &ca, const std::string &dbname,
	const std::string &username, const char *password, std::vector<attrib> &result)
{
  if (ca.empty() || dn.empty())
    return false;
#if MYSQL_CORRECTED
  std::string query = "SELECT usr.dn as username, role, groups.dn as groupname, groups.gid, capability "
    /*    "FROM groups, roles, usr, m, ca, capabilities" */
    "FROM groups, usr, m, ca "
    "left join roles on roles.rid = m.rid "
    "left join capabilities on capabilities.cid = m.cid "
    "WHERE groups.gid = m.gid AND "
    /*    "roles.rid = m.rid AND "
	  "capabilities.cid = m.cid AND "*/
    "usr.uid = m.uid AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn + "\"";
#else
  std::string query = "SELECT m.uid, m.rid, m.gid, m.cid "
    /*    "FROM groups, roles, usr, m, ca, capabilities" */
    "FROM groups, usr, m, ca "
    "WHERE "
    /*    "roles.rid = m.rid AND "
	  "capabilities.cid = m.cid AND "*/
    "usr.uid = m.uid AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn + "\"";
#endif
  return get_group_and_role_real(dn, ca, dbname, username, password, query, result);
}


/*
 * Function:
 *   get_role(dn, ca, role, dbname, username, password)
 *
 * Description:
 *   This function creates a SQL query that retrieves informations about
 *   the groups and roles to which the user belongs, limiting the results
 *   to the role specified.
 * Parameters:
 *   'dn'       - Distinguished Name of the user's certificate.
 *   'ca'       - Distinguished Name of the CA that issued the user's
 *                certificate.
 *   'role'     - The role the user requested.
 *   'username' - The username used for database login.
 *   'password' - The password used for database login.
 *   'dbname'   - The database name.
 *
 * Return:
 *   Success - 
 *     The blob with the requested data.
 *   Failure -
 *     NULL.
 */
bool
get_role(const std::string &dn, const std::string &ca, const char *role,
	 const std::string &dbname, const std::string &username, const char *password,
	 std::vector<attrib> &result)
{
  if (dn.empty() || ca.empty() || !role || dbname.empty() || username.empty() || !password || !acceptable(role))
    return false;

#if MYSQL_CORRECTED
  std::string query = std::string("SELECT usr.dn as username, role, groups.dn as groupname, capability, groups.gid "
    "FROM groups, usr, m, ca "
    "left join roles on roles.rid = m.rid "
    "left join capabilities on capabilities.cid = m.cid "
    "WHERE groups.gid = m.gid AND "
    "usr.uid = m.uid AND "
    "roles.role = \"") + role + "\" AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn + "\"";
#else
  std::string query = std::string("SELECT m.uid, m.rid, m.gid, m.cid "
    "FROM groups, usr, m, ca "
    "WHERE groups.gid = m.gid AND "
    "usr.uid = m.uid AND "
    "roles.rid = m.rid AND "
    "roles.role = \"") + role + "\" AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn + "\"";
#endif

  return get_group_and_role_real(dn, ca, dbname, username, password, query, result);
}

/*
 * Function:
 *   get_group(dn, ca, group, dbname, username, password)
 *
 * Description:
 *   This function creates a SQL query that retrieves informations about
 *   the groups and roles to which the user belongs, limiting the results
 *   to the group specified.
 * Parameters:
 *   'dn'       - Distinguished Name of the user's certificate.
 *   'ca'       - Distinguished Name of the CA that issued the user's
 *                certificate.
 *   'group'    - The group the user requested.
 *   'username' - The username used for database login.
 *   'password' - The password used for database login.
 *   'dbname'   - The database name.
 *
 * Return:
 *   Success - 
 *     The blob with the requested data.
 *   Failure -
 *     NULL.
 */
bool
get_group(const std::string &dn, const std::string &ca, const char *group,
	  const std::string &dbname, const std::string &username, const char *password,
	  std::vector<attrib> &result)
{
  if (dn.empty() || ca.empty() || !group || dbname.empty() || username.empty() || !password || !acceptable(group))
    return false;

#if MYSQL_CORRECTED
  std::string query = std::string("SELECT usr.dn as username, role, groups.dn as groupname, capability, groups.gid "
    "FROM groups, usr, m, ca "
    "left join roles on roles.rid = m.rid "
    "left join capabilities on capabilities.cid = m.cid "
    "WHERE groups.gid = m.gid AND "
    "usr.uid = m.uid AND "
    "groups.dn = \"") + group + "\" AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn +"\"";
#else
  std::string query = std::string("SELECT m.uid, m.rid, m.gid, m.cid "
    "FROM groups, usr, m, ca "
     "WHERE groups.gid = m.gid AND "
    "usr.uid = m.uid AND "
    "groups.dn = \"") + group + "\" AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn +"\"";
#endif

  return get_group_and_role_real(dn, ca, dbname, username, password, query, result);
}

/*
 * Function:
 *   get_group_and_role(dn, ca, message, dbname, username, password)
 *
 * Description:
 *   This function creates a SQL query that retrieves informations about
 *   the groups and roles to which the user belongs, limiting the results
 *   to the group and role specified.
 * Parameters:
 *   'dn'       - Distinguished Name of the user's certificate.
 *   'ca'       - Distinguished Name of the CA that issued the user's
 *                certificate.
 *   'message'  - The group and role the user requested, in the format
 *                "Bgroup:role".
 *   'username' - The username used for database login.
 *   'password' - The password used for database login.
 *   'dbname'   - The database name.
 *
 * Return:
 *   Success - 
 *     The blob with the requested data.
 *   Failure -
 *     NULL.
 */
bool
get_group_and_role(const std::string &dn, const std::string &ca,
		   const char *group, const std::string &dbname,
		   const std::string &username, const char *password,
		   std::vector<attrib> &result)
{
  char *role;

  if (dn.empty() || ca.empty() || !group || dbname.empty() || username.empty() || !password)
    return false;

  char *argument = strdup(group);

  if (!argument || !(role = strchr(argument,':')))
    return false;

  *role++ = '\0';

  if (!acceptable(argument) || !acceptable(role))
    return false;

#if MYSQL_CORRECTED
  std::string query = std::string("SELECT usr.dn as username, role, groups.dn as groupname, capability groups.gid "
    "FROM groups, usr, m, ca "
    "left join roles on roles.rid = m.rid "
    "left join capabilities on capabilities.cid = m.cid "
    "WHERE groups.gid = m.gid AND "
    "usr.uid = m.uid AND "
    "roles.role = \"") + role + "\" AND "
    "groups.dn = \"" + argument + "\" AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn + "\"";
#else
  std::string query = std::string("SELECT m.uid, m.rid, m.gid, m.cid "
    "FROM groups, usr, m, ca "
    "WHERE groups.gid = m.gid AND "
    "roles.rid = m.rid AND "
    "usr.uid = m.uid AND "
    "roles.role = \"") + role + "\" AND "
    "groups.dn = \"" + argument + "\" AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn + "\"";
#endif
  free(argument);
  return get_group_and_role_real(dn, ca, dbname, username, password, query, result);
}

/*
 * Function:
 *   get_group_and_role_real(dbname, username, password, query)
 *
 * Description:
 *   This function executes one of the get_* queries and returns the results
 *   in a blob.
 *
 * Parameters:
 *   'dbname'   - The database name.
 *   'username' - The username used for database login.
 *   'password' - The password used for database login.
 *   'query'    - The query to execute.
 *
 * Return:
 *   Success - 
 *     The blob with the requested data.
 *   Failure -
 *     NULL.
 */
static bool
get_group_and_role_real(const std::string &dn, const std::string &ca,
			const std::string &dbname, const std::string &username,
			const char *password, const std::string &query,
			std::vector<attrib> &results)
{
  std::string res;

#if MYSQL_CORRECTED
  std::string additional = " UNION SELECT usr.dn as username, role, groups.dn as groupname, capability, groups.gid "
    /*    "FROM groups, roles, usr, m, ca, capabilities "*/
    "FROM groups, usr, m, ca "
    "left join roles on roles.rid = m.rid "
    "left join capabilities on capabilities.cid = m.cid "
    "WHERE groups.gid = m.gid AND "
    /*    "roles.rid = m.rid AND " */
    "usr.uid = m.uid AND "
    /*    "capabilities.cid = m.cid AND " */
    "groups.must IS NOT NULL AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn + "\"";
#else
  std::string additional = " UNION SELECT m.uid, m.rid, m.gid, m.cid "
    /*    "FROM groups, roles, usr, m, ca, capabilities "*/
    "FROM groups, usr, m, ca "
    "WHERE groups.gid = m.gid AND "
    /*    "roles.rid = m.rid AND " */
    "usr.uid = m.uid AND "
    /*    "capabilities.cid = m.cid AND " */
    "groups.must IS NOT NULL AND "
    "usr.ca  = ca.cid AND "
    "ca.ca = \"" + ca + "\" AND "
    "usr.dn = \"" + dn + "\"";
#endif
  if (dbname.empty() || username.empty() || !password || query.empty())
    return false;

  std::string complete_query = query + additional;

  try {
    sql::interface db;
    db.connect(dbname.c_str(), "localhost", username.c_str(), password);
#ifndef MYSQL_CORRECTED
    sql::interface db2;
    db2.connect(dbname.c_str(), "localhost", username.c_str(), password);
#endif
    //    db.lock("LOCK TABLE groups READ, roles READ, usr READ, m READ, ca READ, capabilities READ, groups as g1 READ");

    sql::query q(db);
    q << "SET AUTOCOMMIT = 0;";
    q << "SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;";
    datahash map;

#if MYSQL_CORRECTED
    bool ok = get_group_and_role_real0(db, dn, ca, complete_query, results, map);
#else
    bool ok = get_group_and_role_real0(db, db2, dn, ca, complete_query, results, map);
#endif
    //    db.unlock();
    q << "ROLLBACK;";
    return ok;
  }
  catch(...) {}

  return false;

}


static bool
#ifdef MYSQL_CORRECTED
get_group_and_role_real0(sql::interface &db,
			 const std::string &dn, const std::string &ca,
			 const std::string &query,
			 std::vector<attrib> &results,
			 datahash &map)
#else
get_group_and_role_real0(sql::interface &db, sql::interface &db2,
			 const std::string &dn, const std::string &ca,
                         const std::string &query, 
			 std::vector<attrib> &results,
			 datahash &map)
#endif
{
  std::string res;
  bool flag = false;

  if (query.empty())
    return false;

  try {
    sql::query q(db);

    cerr << "QUERY: " << query << endl;

    q << query.c_str();

    sql::iterator i = q.begin(), j = q.end();

    std::string group;

    vector<std::string> v;

    while (!(i == j)) {
      attrib rec;
      group = (&i)["gid"];
      
      std::string record;

#if MYSQL_CORRECTED
      rec.group = (&i)["groupname"];
      rec.role = (&i)["role"];
      rec.cap = (&i)["capability"];
      record = "GROUP: " + rec.group + "\nROLE: " + rec.role + "\nCAP: " + rec.cap + "\n";
#else
      std::string role  = (&i)["rid"];
      std::string cap   = (&i)["cid"];
      cerr << "Examining: " << group << ", " << role << ", " << cap << endl;
      std::string newq  = "SELECT dn FROM groups WHERE gid = " + group + ";";
      if (group != "NULL")
	cerr << simple_query0(db2, newq, "dn", group) << endl << flush;
      if (role != "NULL")
	simple_query0(db2, 
		      "SELECT role FROM roles WHERE rid = " +role+";",
		      "role",role);
      if (cap != "NULL")
	simple_query0(db2, 
		      "SELECT capability FROM capabilities WHERE cid = " +cap+";",
		      "capability",cap);

      rec.group = group;
      rec.role = role;
      rec.cap = cap;
      record = "GROUP: " + group + "\nROLE: " + role + "\nCAP: " + cap + "\n";
#endif
      if (map.find(record) == map.end()) {
	map.insert(record);
	v.push_back(group);
      }
      results.push_back(rec);
      ++i;
    }
    for (unsigned int i=0; i < v.size(); i++) {
#if MYSQL_CORRECTED
      std::string newq = "SELECT usr.dn as username, role, groups.dn as groupname, capability, groups.gid "
	/*	"FROM groups, roles, usr, m, ca, capabilities, groups as g1 "*/
	"FROM groups, usr, m, ca, groups as g1 "
	"left join roles on roles.rid = m.rid "
	"left join capabilities on capabilities.cid = m.cid "
	"WHERE g1.parent = groups.gid AND "
	"groups.parent = m.gid AND "
	/*	"roles.rid = m.rid AND "*/
	"usr.uid = m.uid AND "
	/*	"capabilities.cid = m.cid AND "*/
	"g1.gid = " + v[i] + " AND "
	"usr.ca  = ca.cid AND "
	"ca.ca = \"" + ca + "\" AND "
	"usr.dn = \"" + dn + "\"";

      bool val = get_group_and_role_real0(db, dn, ca, newq, results, map);
#else
      std::string newq = "SELECT m.uid, m.rid, m.gid, m.cid "
	/*	"FROM groups, roles, usr, m, ca, capabilities, groups as g1 "*/
	"FROM groups, usr, m, ca, groups as g1 "
	"WHERE g1.parent = groups.gid AND "
	"groups.parent = m.gid AND "
	/*	"roles.rid = m.rid AND "*/
	"usr.uid = m.uid AND "
	/*	"capabilities.cid = m.cid AND "*/
	"g1.gid = \"" + v[i] + "\" AND "
	"usr.ca  = ca.cid AND "
	"ca.ca = \"" + ca + "\" AND "
	"usr.dn = \"" + dn + "\"";
      bool val = get_group_and_role_real0(db, db2, dn, ca, newq, results, map);
#endif

      if (!val)
	return false;
    }
    flag = true;
  }
  catch (...) {}

  return flag;
}

/*
 * Function:
 *   get_fields(dbname, username, password, query)
 *
 * Description:
 *   This function executes the query given in its 'query' argument,
 *   and then encodes the resulting table in blob. See special() for the
 *   exact format of the blob.
 *
 * Parameters:
 *   'dbname'   - The database name.
 *   'username' - The username used for database login.
 *   'password' - The password used for database login.
 *   'query'    - The query to execute.
 *
 * Return:
 *   Success - 
 *     The blob with the requested data.
 *   Failure -
 *     NULL.
 */
static bool
get_fields(const std::string &dbname, const std::string &username, const char *password,
	   const std::string &query, std::string &result)
{
  std::string res;

  if (dbname.empty() || username.empty() || !password || query.empty())
    return false;

  try {
    sql::interface db;
    db.connect(dbname.c_str(),"localhost", username.c_str(), password);
    sql::query q(db);

    q << query.c_str();
    //    q.exec();

    sql::iterator i = q.begin(), j = q.end();

    while (!(i == j)) {
      for (int index = 0; index < (&i).size(); index++)
	res += i.name(index) + ": " + (&i)[index] + "\n";
      ++i;
    }

    result += res;
    return true;
  }
  catch (...) {}
  return false;
}

/*
 * Function:
 *   simple_query(dbname, username, password, query, result)
 *
 * Description:
 *   This function executes the query given in its 'query' argument,
 *   and then returns the content of the 'result' field of the first row.
 *
 * Parameters:
 *   'dbname'    - The database name.
 *   'username'  - The username used for database login.
 *   'password'  - The password used for database login.
 *   'query'     - The query to execute.
 *   'fieldname' - The name of the field you are interested in.
 *
 * Return:
 *   Success - 
 *     A std::string with the requested data.
 *   Failure -
 *     NULL.
 */
static bool
simple_query(const std::string &dbname, const std::string &username,
	     const char *password, const std::string &query,
	     const char *fieldname, std::string &result)
{
  if (dbname.empty() || username.empty() || !password || 
      !fieldname || query.empty())
    return false;

  try {
    sql::interface db;
    db.connect(dbname.c_str(),"localhost", username.c_str(), password);

    return simple_query0(db, query, fieldname, result);
  }
  catch (...) {}
  return false;
}

static bool
simple_query0(sql::interface &db, const std::string &qq,
	     const char *fieldname, std::string &result)
{
  if (!fieldname || qq.empty())
    return false;

  cerr <<qq <<endl<< flush;
  try {
    sql::query q(db);

    q << qq.c_str();
    
    sql::iterator i = q.begin(), j = q.end();

    if (!(i == j)) {
      result = (&i)[fieldname];
      return true;
    }
  }
  catch (...) {}
  return false;
}
