#ifndef GRID_MANAGER_RUN_H
#define GRID_MANAGER_RUN_H

#include "../std.h"
#include "../jobs/users.h"
#include "../jobs/states.h"
#include <sys/resource.h>
#include <sys/wait.h>
#include <globus_common.h>

extern char** environ;

class RunParallel;
class RunElementPointer;


class RunPlugin {
 private:
  list<string> args_;
  string lib;
  string stdin_;
  string stdout_;
  string stderr_;
  int timeout_;
  int result_;
  void set(const string& cmd);
  void set(char const * const * args);
 public:
  typedef void (*substitute_t)(string& str,void* arg);
  typedef int (*lib_plugin_t)(...);
  RunPlugin(void):timeout_(10),result_(0) { };
  RunPlugin(const string& cmd):timeout_(10),result_(0) { set(cmd); };
  RunPlugin(char const * const * args):timeout_(10),result_(0) { set(args); };
  void operator=(const string& cmd) { set(cmd); };
  void operator=(char const * const * args) { set(args); };
  bool run(void);
  bool run(substitute_t subst,void* arg);
  int result(void) const { return result_; };
  void timeout(int t) { timeout_=t; };
  void stdin(const string& s) { stdin_=s; };
  const string& stdout(void) const { return stdout_; };
  const string& stderr(void) const { return stderr_; };
  operator bool(void) const { return (args_.size() > 0); };
};

class RunPlugins {
 private:
  list<RunPlugin*> plugins_;
  int result_;
 public:
  RunPlugins(void):result_(0) { };
  void add(const string& cmd);
  bool run(void);
  bool run(RunPlugin::substitute_t subst,void* arg);
  int result(void) { return result_; };
};

class RunElement {
 friend class RunParallel;
 friend class RunElementPointer;
 private:
  int pid;
  int exit_code;
  bool released;
  RunElement* next;
  RunElement* next_u;
 public:
  void kill(int signum = SIGTERM) {
    int tmp_pid=pid; if(tmp_pid != -1) ::kill(tmp_pid,signum);
  };
  int get_exit_code(void) { return exit_code; };
  RunElement() { pid=0; exit_code=-1; released=false; next=NULL; next_u=NULL; };
  RunElement(const RunElement &re) { 
    pid=re.pid; exit_code=re.exit_code; released=re.released;
    next=NULL; next_u=NULL;
  };
  ~RunElement() { };
};

class RunElementPointer {
 friend class RunParallel;
 private:
  RunElement* prev;
  RunElement* cur;
 public:
  RunElementPointer(void) { cur=NULL; prev=NULL; };
  RunElementPointer(RunElement* re) { cur=re; prev=NULL; };
  ~RunElementPointer(void) { };
  bool is_end(void) { return (cur==NULL); };
  void go_next(void) { if(cur) { prev=cur; cur=cur->next; }; };
  void go_next_u(void) { if(cur) { prev=cur; cur=cur->next_u; };
  };
  void operator=(RunElement* re) { prev=NULL; cur=re; };
  RunElement* operator->() { return cur; };
};

class RunParallel {
 private:
  bool initialized;
  static bool hup_detected;
  static RunElement dumb_re;
  static bool in_signal;
  static globus_cond_t *cond;
  static struct sigaction old_sig_chld; 
  static struct sigaction old_sig_hup; 
  static struct sigaction old_sig_term; 
  static bool old_sig_chld_inited; 
  static bool old_sig_hup_inited; 
  static bool old_sig_term_inited; 
  static void sig_hup(int signum,siginfo_t *info,void* arg);
  static void sig_chld(int signum,siginfo_t *info,void* arg);
  static void sig_term(int signum,siginfo_t *info,void* arg);
  static RunElement* begin;
  static RunElement* begin_u;
  bool init(void);
  void deinit(void);
 public:
  static bool run(JobUser& user,const char* jobid,char *const args[],RunElement**,bool su = true,bool job_proxy = true, RunPlugin* cred = NULL, RunPlugin::substitute_t subst = NULL, void* subst_arg = NULL);
  static bool run(JobUser& user,const JobDescription& desc,char *const args[],RunElement**,bool su = true);
  static bool plain_run_piped(char *const args[],const string *Din,string *Dout,string *Derr,unsigned int timeout,int* result);
  static void release(RunElement* re);
  static RunElement* add_handled(RunElement* re);
  bool is_initialized(void) { return initialized; };
  bool was_hup(void);
  RunParallel(globus_cond_t *cond = NULL);
  ~RunParallel();
};

void free_args(char** args);
char** string_to_args(const string& command);

#endif // GRID_MANAGER_RUN_H

